/******************************************************************************
 * Wiznet W5100/W3150 Driver.
 * Copyright Andreas Kaiser 2007.
 ******************************************************************************/

#include "config.h"
#include "common.h"
#include "w5100.h"
#include "w5100-common.h"

#define BasePort  10000

uint16_t          w5port = 0;		// used in port autoassignment, see xx_open()
struct W5Socket_s w5so[4];

//-----------------------------------------------------------------------------

unsigned
tcpip_load16(uint8_t *p)
{
    return p[0] << 8 | p[1];
}

unsigned
tcpip_load32(uint8_t *p)
{
    return (uint32_t)tcpip_load16(p+0) << 16 | tcpip_load16(p+2);
}

void
tcpip_store16(uint8_t *p, unsigned v)
{
    p[0] = v >> 8;
    p[1] = v;
}

void
tcpip_store32(uint8_t *p, uint32_t v)
{
    tcpip_store16(p+0, v >> 16);
    tcpip_store16(p+2, v);
}

//-----------------------------------------------------------------------------

void
tcpip_paddr(const char *s, uint32_t addr)
{
    rprintf("%s%u.%u.%u.%u", s, IPBYTE0(addr), IPBYTE1(addr), IPBYTE2(addr), IPBYTE3(addr));
}

const char *
tcpip_error(int status)
{
    const char *names[] = {
	/*0*/	"Closed",
	/*1*/	"ARP",
	/*2*/	"UDP/Raw",
	/*3*/	"SockInit",
	/*4*/	"SockListen",
	/*5*/	"SynSent",
	/*6*/	"SynRecv",
	/*7*/	"Established",
	/*8*/	"FinWait",
	/*9*/	"9?",
	/*A*/	"Closing",
	/*B*/	"TimeWait",
	/*C*/	"CloseWait",
	/*D*/	"LastAck",
	/*E*/	"E?",
	/*F*/	"PPPoE",
    };
    if (status == -1)
	return "Busy";
    if (status < 0)
	status = -status - 0x100;
    return names[status & 0x0F];
}

int
w5status(unsigned so)
{
    return w5regread(W5sSR(so), 1);
}

int
tcpip_status(unsigned so)
{
    return W5STATUS(w5status(so));
}

void
tcpip_dump(unsigned so, const char *s)
{
    if (s)
	W5DEBUG(1, "%s#%u:", s?s:"", so);
    W5DEBUG(1, " MR=%02lX", so, w5regread(W5sMR(so),    1));
    W5DEBUG(1, " CR=%02lX",     w5regread(W5sCR(so),    1));
    W5DEBUG(1, " IR=%02lX",     w5regread(W5sIR(so),    1));
    W5DEBUG(1, " SR=%02lX",     w5regread(W5sSR(so),    1));
    W5DEBUG(1, " SPORT=%ld",    w5regread(W5sSPORT(so), 2));
    W5PADDR(1, " DIP=",         w5regread(W5sDIPR(so),  4));
    W5DEBUG(1, ":%lu",          w5regread(W5sDPORT(so), 2));
    W5DEBUG(1, " TOS=%lu",      w5regread(W5sTOS(so),   1));
    W5DEBUG(1, " TTL=%lu",      w5regread(W5sTTL(so),   1));
    W5DEBUG(1, " MSS=%lu\n",    w5regread(W5sMSSR(so),  2));
    W5DEBUG(1, "\tTXFSR=%ld",   w5regread(W5sTXFSR(so), 2));
    W5DEBUG(1, " TXRD=%04lX",   w5regread(W5sTXRD(so),  2));
    W5DEBUG(1, " TXWR=%04lX",   w5regread(W5sTXWR(so),  2));
    W5DEBUG(1, " RXRSR=%ld",    w5regread(W5sRXRSR(so), 2));
    W5DEBUG(1, " RXRD=%04lX\n", w5regread(W5sRXRD(so),  2));
}

void
tcpip_dumpall(void)
{
#if W5DEBUGLVL
    uint8_t mac[6];
    W5DEBUG(1, "MR=%02X", w5regread(W5MR, 1));
    W5DEBUG(1, " IR=%02X", w5regread(W5IR, 1));
    W5DEBUG(1, " IMR=%02X\n", w5regread(W5IMR, 1));
    w5blkread(W5SHAR, mac, 6);
    W5DEBUG(1, "\tMAC=%02X:%02X:%02X:%02X:%02X:%02X", mac[0],mac[1],mac[2],mac[3],mac[4],mac[5]);
    W5PADDR(1, " IP=", w5regread(W5SIPR, 4));
    W5PADDR(1, " MASK=", w5regread(W5SUBR, 4));
    W5PADDR(1, " GW=", w5regread(W5GAR,  4));
    W5DEBUG(1, "\n");
    for (unsigned so = 0; so < 4; ++so)
	tcpip_dump(so, 0);
#endif
}

//-----------------------------------------------------------------------------

void
w5command(unsigned so, unsigned cmd)
{
    w5regwrite(W5sCR(so), cmd, 1);
}

void
w5commandWait(unsigned so, unsigned cmd)
{
    unsigned val;
    w5command(so, cmd);
    while ((val = w5regread(W5sCR(so), 1)) != 0) {
	W5DEBUG(3, "tcpip:wait #%u cmd=%02X reg=%02X\n", so, cmd, val);
    }
}

//-----------------------------------------------------------------------------

int
tcpip_reset(void)
{
    w5init();
    W5DEBUG(1, "tcpip:reset\n");
    w5regwrite(W5MR, W5MR_SWReset, 1);
    return 0;
}

void
tcpip_macaddr(uint8_t mac[6])
{
    W5DEBUG(1, "tcpip:macaddr %02X:%02X:%02X:%02X:%02X:%02X\n",
	    mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
    w5blkwrite(W5SHAR, mac, 6);
}

void
tcpip_ipaddr(uint32_t ipaddr, uint32_t ipmask, uint32_t gwaddr)
{
    W5PADDR(1, "tcpip:ipaddr addr=", ipaddr);
    W5PADDR(1, ", mask=", ipmask);
    W5PADDR(1, ", gw=", gwaddr);
    W5DEBUG(1, "\n");
    w5regwrite(W5SIPR, ipaddr, 4);
    w5regwrite(W5SUBR, ipmask, 4);
    w5regwrite(W5GAR,  gwaddr, 4);
}

void
tcpip_timeout(unsigned initial, unsigned retries)
{
    W5DEBUG(1, "tcpip:timeout initial=%u retries=%u\n", initial, retries);
    w5regwrite(W5RTR, initial, 2);
    w5regwrite(W5RCR, retries, 1);
}

void
tcpip_start(unsigned rxmem, unsigned txmem)
{
    W5DEBUG(1, "tcpip:init rxmem=%02X txmem=%02X\n", rxmem, txmem);

    w5regwrite(W5RMSR, rxmem, 1);
    w5regwrite(W5TMSR, txmem, 1);

    w5regwrite(W5IMR, W5IR_Conflict | W5IR_Unreach | 0x0F, 1);

    unsigned rbase = W5RXBUF;
    unsigned tbase = W5TXBUF;
    for (int i = 0; i < 4; ++i) {
	unsigned size;

	w5so[i].status = 0;

	size = 1024 << (txmem & 3);
	txmem >>= 2;
	if (tbase + size <= W5TXBUF_END) {
	    w5so[i].tbase = tbase;
	    w5so[i].tsize = size;
	    tbase += size;
	}

	size = 1024 << (rxmem & 3);
	rxmem >>= 2;
	if (rbase + size <= W5RXBUF_END) {
	    w5so[i].rbase = rbase;
	    w5so[i].rsize = size;
	    rbase += size;
	}

	W5DEBUG(1, "\t%d: R=%04X-%04X T=%04X-%04X\n", i,
	      w5so[i].rbase, (w5so[i].rbase + w5so[i].rsize-1) & 0xFFFF,
	      w5so[i].tbase, (w5so[i].tbase + w5so[i].tsize-1) & 0xFFFF);
    }
}

//-----------------------------------------------------------------------------

int
tcpip_local(unsigned so, unsigned localport)
{
    if (localport == 0)
	localport = BasePort + so * 1024 + (w5port++ % 1024);
    w5regwrite(W5sSPORT(so), localport, 2);
    return 0;
}

int
tcpip_open(unsigned so, unsigned mode)
{
    w5regwrite(W5sMR(so), mode, 1);
    w5regwrite(W5sTOS(so), 0, 1);
    w5regwrite(W5sTTL(so), 30, 1);
    w5commandWait(so, W5sCR_Open);
    return 0;
}

int
tcpip_close(unsigned so)
{
    W5DEBUG(1, "tcpip:close #%u close\n", so);
    w5command(so, W5sCR_Close);
    return tcpip_closed(so);
}

int
tcpip_closed(unsigned so)
{
    int status = w5status(so);
    W5DEBUG(2, "tcpip:closed #%u status=%s\n", so, tcpip_error(status));
    if (status == W5sSR_Closed)
	return 0;
    if (status >= W5sSR_FinWait && status <= W5sSR_LastAck)
	return status;
    return W5STATUS(status);
}

//-----------------------------------------------------------------------------

int
tcpip_write_avail(unsigned so)
{
    unsigned n = w5regread(W5sTXFSR(so), 2);
#if 0
    unsigned n2;;
    while ((n2 = w5regread(W5sTXFSR(so), 2)) != n) {
	W5DEBUG(1, "tcpip:write_avail #%u %u\n", n, n2);
	n = n2;
    }
#endif
    W5DEBUG(3, "tcpip:write_avail #%u available=%d\n", so, n);
    return n;
}

void
tcpip_write(unsigned so, unsigned offset, const uint8_t *data, int nb)
{
    unsigned size = w5so[so].tsize;
    offset &= size - 1;
    W5DEBUG(3, "tcpip:write #%u offset=%04X size=%d\n", so, offset, nb);
    if (offset + nb > size) {
	size -= offset;
	w5blkwrite(w5so[so].tbase + offset, data, size);
	w5blkwrite(w5so[so].tbase, (char *)data + size, nb - size);
    } else
	w5blkwrite(w5so[so].tbase + offset, data, nb);
}

//-----------------------------------------------------------------------------

int
tcpip_read_avail(unsigned so)
{
    unsigned n = w5regread(W5sRXRSR(so), 2), n2;
#if 0
    unsigned n2;
    while ((n2 = w5regread(W5sRXRSR(so), 2)) != n) {
	W5DEBUG(1, "tcpip:read_avail #%u %u\n", n, n2);
	n = n2;
    }
#endif
    W5DEBUG(3, "tcpip:read_avail #%u available=%d\n", so, n);
    return n;
}

void
tcpip_read(unsigned so, unsigned offset, uint8_t *data, int nb)
{
    unsigned size = w5so[so].rsize;
    offset &= size - 1;
    W5DEBUG(3, "tcpip:read #%u offset=%04X size=%d\n", so, offset, nb);
    if (offset + nb > size) {
	size -= offset;
	w5blkread(w5so[so].rbase + offset, data, size);
	w5blkread(w5so[so].rbase, data + size, nb - size);
    } else
	w5blkread(w5so[so].rbase + offset, data, nb);
}

//-----------------------------------------------------------------------------

void
w5interrupt(void)
{
    unsigned ir = w5regread(W5IR, 1);
    W5DEBUG(2, "[*W5IRQ:%02X", ir);
    w5regwrite(W5IR, 0xFF, 1);
    for (unsigned so = 0; so < 4; ++so) {
	if (ir & (1 << so)) {
	    unsigned sir = w5regread(W5sIR(so), 1);
	    w5regwrite(W5sIR(so), sir, 1);
	    W5DEBUG(2, ",%u:%02X", so, sir);
	}
    }
    W5DEBUG(2, "*]");
}
