/******************************************************************************
 * Wiznet W5100/W3150 Driver.
 * Copyright Andreas Kaiser 2007.
 ******************************************************************************/

#ifndef W5100_H_
#define W5100_H_

/*
 * Utilitities.
 */

#define IPBYTE0(u32)	((u32)>>24&0xFF)
#define IPBYTE1(u32)	((u32)>>16&0xFF)
#define IPBYTE2(u32)	((u32)>>8&0xFF)
#define IPBYTE3(u32)	((u32)&0xFF)

#define IPADDR(u0,u1,u2,u3) ((uint32_t)(u0)<<24|(uint32_t)(u1)<<16|(uint32_t)(u2)<<8|(uint32_t)(u3))

/******************************************************************************
 * Return codes:
 *	>= 0:	success
 *	-1:	busy
 *	-1XX:	error, XX=status
 */

void		tcpip_paddr  (const char *s, uint32_t addr);
int		tcpip_status (unsigned so);
const char *	tcpip_error  (int status);

void		tcpip_dump (unsigned so, const char *s);
void		tcpip_dumpall (void);

/******************************************************************************
 * Initialization.
 */

// reset device - return 0 on success, some status code on failure
int	tcpip_reset (void);
// then setup addresses
void	tcpip_macaddr (uint8_t mac[6]);
void	tcpip_ipaddr  (uint32_t ipaddr, uint32_t ipmask, uint32_t gwaddr);
void	tcpip_timeout (unsigned initial, unsigned retries);
// then initialize
void	tcpip_start (unsigned rxmem, unsigned txmem);

/******************************************************************************
 * Internal use.
 */

int	tcpip_local  (unsigned so, unsigned localport);
int	tcpip_open   (unsigned so, unsigned mode);
int	tcpip_close  (unsigned so);
int	tcpip_closed (unsigned so);

int	tcpip_write_avail (unsigned so);
void	tcpip_write (unsigned so, unsigned offset, const uint8_t *data, int nb);

int	tcpip_read_avail (unsigned so);
void	tcpip_read (unsigned so, unsigned offset, uint8_t *data, int nb);

/******************************************************************************
 * TCP Connections.
 *
 * Usually return >=0 on success, -1 on busy, <0 on error
 */

// initialize stream socket, choose default if localport==0, blocking
int	tcp_open (unsigned so, unsigned localport);

// initiate connection to remote (active mode), return >0 when busy
int	tcp_connect (unsigned so, uint32_t remoteaddr, unsigned remoteport);

// accept connection by remote (passive mode)
int	tcp_listen (unsigned so);

// return 0 when we have a established connection, >1 when busy
int	tcp_established (unsigned so);
uint32_t tcp_remote_addr (unsigned so);
unsigned tcp_remote_port (unsigned so);

// return tx buffer space available
int	tcp_send_avail (unsigned so);

// send data, nonblocking
//	return no of bytes enqueued
int	tcp_send (unsigned so, const void *data, int size);

// receiver data available at least <size> bytes
int	tcp_recv_avail (unsigned so, int size);

// receive data, nonblocking
//	return no of bytes received
int	tcp_recv (unsigned so, void *data, int size);

// disconnect&close
int	tcp_close  (unsigned so);
int	tcp_cancel (unsigned so);	// drop data (reinit socket), see tcp_closed() for rc
#define	tcp_closed(so) tcpip_close(so)

/******************************************************************************
 * UDP Connections.
 *
 * Usually return >=0 on success, -1 on busy, <0 on error
 */

// initialize datagram socket, choose default if localport==0, blocking
int	udp_open (unsigned so, unsigned localport);

// return tx buffer space available, <0 on error
//	return -1 when prior transmission is still pending
int	udp_send_avail (unsigned so);

// send data, nonblocking
//	return no of bytes enqueued
//	return 0 when busy
//	return <0 on error
int	udp_send (unsigned so, void *data, int size,
		  uint32_t remoteaddr, unsigned remoteport);

// return size of next datagram available
int	udp_recv_avail (unsigned so);

// receive data, nonblocking
//	return no of bytes received
//	return 0 when busy
//	return <0 on error
int	udp_recv (unsigned so, void *data, int size,
		  uint32_t *remoteaddr, uint16_t *remoteport);

// close
#define	udp_close(so)  tcpip_close(so)
#define	udp_closed(so) tcpip_closed(so)

/******************************************************************************
 * Utilities.
 */

unsigned	tcpip_load16 (uint8_t *);
unsigned	tcpip_load32 (uint8_t *);
void		tcpip_store16 (uint8_t *, unsigned);
void		tcpip_store32 (uint8_t *, uint32_t);

#endif //W5100_H_
